package sensorGlove;

import gnu.io.*; // RXTX

import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.nio.ByteBuffer;
import java.nio.CharBuffer;
import java.util.Enumeration;
import java.util.TooManyListenersException;

// GloveBT stands for Sensor Glove with a Bluetooth Connection //
// features: IMU, battery level, flex sensors and vibration motors //
// Author: Elmar Rueckert, rueckert.elmar@gmail.com //
// Date: 28th January, 2016 //

public class GloveBT extends Thread {

	private final PropertyChangeSupport pcs = new PropertyChangeSupport(this);
	
	public void addPropertyChangeListener(PropertyChangeListener listener) {
        this.pcs.addPropertyChangeListener(listener);
    }

    public void removePropertyChangeListener(PropertyChangeListener listener) {
        this.pcs.removePropertyChangeListener(listener);
    }

    public static class IncomingData {
    	//Sensor recordings  (if not stated differently, all values are within [0,255] )
		//Flexsensors for proximal and distal joints
		//flex11, flex12, ... flex51, flex52
		int[] flexValues = new int[10];
		//IMU
		//ax,ay,az; gx,gy,gz; mx, my, mz;
		int[] imuACCValues = new int[3];
		int[] imuGYROValues = new int[3];
		int[] imuMAGNValues = new int[3];
		//Batterie charge status
		//batt			//Batterie-Spannung zwischen 6V und 8,4V. D.h. Pro Zelle (die hier gemessen wird) 3V und 3,7V = 630 bis 855 /1023
		int batteryChargeLevel;
		long timeInMilliSeconds;
        
		public int[] getFlexValues(){
        	return this.flexValues;
        }
		
        public int[] getImuACCValues(){
        	return this.imuACCValues;
        }
        public int[] getImuGYROValues(){
        	return this.imuGYROValues;
        }
        public int[] getImuMAGNValues(){
        	return this.imuMAGNValues;
        }
        
        public int getBatteryChargeLevel(){
        	return this.batteryChargeLevel;
        }
        
        public long getTimeInMilliSeconds(){
        	return this.timeInMilliSeconds;
        }
        
        private void parseGroupValues(String type, String number, String value){

        	//System.out.println("type " + type);
        	int numberInt, valueInt;
        	try {
        		numberInt = Integer.parseInt(number);
        		valueInt = Integer.parseInt(value);
        	}catch(NumberFormatException ex){
        		System.out.println("error, could not parse integers: type " + type + " number "  + number + " value " + value);
        		return;
        	}
        	
//        	if(type.equals("acc")) {
//        		System.out.println("acc received " + value);
//        		//this.imuACCValues[numberInt-1] = valueInt;
//        		
//        	}
        	
        	switch(type){
        	
        	case "acc":
        		this.imuACCValues[numberInt-1] = valueInt;
        		break;
        		
        	case "gyr":
        		this.imuGYROValues[numberInt-1] = valueInt;
        		break;
        		
        	case "mag":
        		this.imuMAGNValues[numberInt-1] = valueInt;
        		break;
        		
        	case "fle":
        		switch(number){
        			case "11": this.flexValues[0] = valueInt; break;
        			case "12": this.flexValues[1] = valueInt; break;
        			case "21": this.flexValues[2] = valueInt; break;
        			case "22": this.flexValues[3] = valueInt; break;
        			case "31": this.flexValues[4] = valueInt; break;
        			case "32": this.flexValues[5] = valueInt; break;
        			case "41": this.flexValues[6] = valueInt; break;
        			case "42": this.flexValues[7] = valueInt; break;
        			case "51": this.flexValues[8] = valueInt; break;
        			case "52": this.flexValues[9] = valueInt; break;
        			default: break;
        		}
        		break;
        		
        	case "bat":
        		this.batteryChargeLevel = valueInt;
        		break;
        	
        	case "mil":
        		this.timeInMilliSeconds = Long.parseLong(value);
        		break;
        		
    		default:
    			throw new IllegalArgumentException("Invalid command in the transmission: " + type);//
    			//System.out.println("____");
        	}
        }
        
//        public void setDataFromBuffer(byte[] data, int idxStart, int numBytesPerTransmission){
//        	String str = new String(data,idxStart,numBytesPerTransmission);
//        	try{
//	        	str = str.replaceAll("\\s","");
//	        	
//	        	String[] attr = str.split(";");
//	        	int i;
//	        	for(i=0; i< attr.length; i++){
////	        		System.out.println("attr " + attr[i]);
//	        		if(attr[i].length() > 3+3){//3 for the command and 3 for at least one number
//	        			parseGroupValues(attr[i]);
//	        		}
//	        	}
//        	}catch(Exception e){
//        		System.out.println("error " + str);
//        	}
//        }
        
    }
    
	CommPortIdentifier serialPortId;
	Enumeration<?> enumComm;
	SerialPort serialPort;
	InputStream inputStream;
	OutputStream outputStream;
	Boolean serialPortOpen = false;
	IncomingData inData;

	int baudrate = 115200;
	int dataBits = SerialPort.DATABITS_8;
	int stopBits = SerialPort.STOPBITS_1;
	int parity = SerialPort.PARITY_NONE;
	int flowControl = SerialPort.FLOWCONTROL_NONE;
	
	String portName = "";
	
	public GloveBT(String portName){
		this.portName = portName;
		this.inData = new IncomingData();
	}
	
	public void open(){
		super.start();
	}
	
	@Override
    public void run(){
		openSerialPort(portName);
    }
    
	protected boolean openSerialPort(String portName){
		Boolean foundPort = false;
		if (serialPortOpen != false) {
			System.out.println("Port already open!");
			return false;
		}
		System.out.println("Opening serialport");
		enumComm = CommPortIdentifier.getPortIdentifiers();
		while(enumComm.hasMoreElements()) {
			serialPortId = (CommPortIdentifier) enumComm.nextElement();
			if (portName.contentEquals(serialPortId.getName())) {
				foundPort = true;
				break;
			}
		}
		
		if (foundPort != true) {
			System.out.println("Could not find port: " + portName);
			return false;
		}
		
		try {
			serialPort = (SerialPort) serialPortId.open("Öffnen und Senden", 500);
		} catch (PortInUseException e) {
			System.out.println("Port already in use.");
		}

		try {
			outputStream = serialPort.getOutputStream();
		} catch (IOException e) {
			System.out.println("Can not access output stream.");
		}

		try {
			inputStream = serialPort.getInputStream();
		} catch (IOException e) {
			System.out.println("Can not access input stream.");
		}
		try {
			serialPort.addEventListener(new serialPortEventListener());
		} catch (TooManyListenersException e) {
			System.out.println("TooManyListenersException");
		}
		serialPort.notifyOnDataAvailable(true);
		try {
			serialPort.setSerialPortParams(baudrate, dataBits, stopBits, parity);
			serialPort.setFlowControlMode(flowControl);
		} catch(UnsupportedCommOperationException e) {
			System.out.println("Could not set port parameters");
		}
		
		serialPortOpen = true;
		return true;
	}

	public void close(){
		if ( serialPortOpen == true) {
			System.out.println("Closing serialport");
			setMotorValues(new int[]{0,0,0,0,0});
			serialPort.close();
			serialPortOpen = false;
		} else {
			System.out.println("Serialport already closed");
		}
	}
	
	protected void serialPortDataReady() {
		int data;
		boolean found = false;
		
		try {
			byte[] buffer = new byte[50];
	        int len = 0;
            while ( ( data = inputStream.read()) > -1 )
            {
                if ( data == '\r' || data == '\n' ) {
                	found = true;
                    break;
                }
                buffer[len++] = (byte) data;
            }
            if(len < 7){ return; }
            
            if(found){
            	String type, number, value, tmpATTR;
            	tmpATTR = new String(buffer,0,len);
            	
                type = new String(buffer,0,3);
                number = new String(buffer,3,2);
                value = new String(buffer,6,len-6);

                this.inData.parseGroupValues(type, number, value);
            	
				this.pcs.firePropertyChange("gloveFlexValues", 0, inData);
			}
			
//			byte[] data = new byte[500];
//			int num = 0;
//			while(inputStream.available() > 0) {
//				num = inputStream.read(data, 0, data.length);
//			}
//
//			boolean found = false;
//			
//			int idx = 0;
//			int numBytesPerTransmission = 40;//40;//TODO
//			for( idx = 0; idx < (num-numBytesPerTransmission); idx ++){
//				if(data[idx] == 59){ //;
//					inData.setDataFromBuffer(data, idx+1, num);
//					found = true;
//					break;
//				}
//			}
//			
//			if(found){
//				this.pcs.firePropertyChange("gloveFlexValues", 0, inData);
//			}
			
		} catch (IOException e) {
			System.out.println("Could not read data from serialport");
		}
	}
	
	public void setMotorValues(int values[]){
		if (serialPortOpen != true)
			return;
		try {
			//Old Sensor glove version
			//outputStream.write(("w,"+values[0]+","+values[1]+","+values[2]+","+values[3]+","+values[4]+"x").getBytes());
			//New sensor glove version
			//TODO this does not work yet
			outputStream.write(("fb1:"+values[0]+";fb2:"+values[1]+";fb3:"+values[2]+";fb4:"+values[3]+";fb5:"+values[4]+";").getBytes());
		} catch (IOException e) {
			System.out.println("Could not send data to serialport");
		}
	}
	

	class serialPortEventListener implements SerialPortEventListener {
		public void serialEvent(SerialPortEvent event) {
			switch (event.getEventType()) {
			case SerialPortEvent.DATA_AVAILABLE:
				serialPortDataReady();
				break;
			case SerialPortEvent.BI:
				break;
			case SerialPortEvent.CD:
				break;
			case SerialPortEvent.CTS:
				break;
			case SerialPortEvent.DSR:
				break;
			case SerialPortEvent.FE:
				break;
			case SerialPortEvent.OUTPUT_BUFFER_EMPTY:
				break;
			case SerialPortEvent.PE:
				break;
			case SerialPortEvent.RI:
				break;
			default:
			}
		}
	}
	
}
